import multiprocessing as mp

from docarray import Document, DocumentArray
from tqdm import tqdm

# $ pip uninstall jina
# $ pip uninstall docarray
# $ pip install jina==2.6.4
# $ mv .venv/lib/python3.8/site-packages/docaarray .venv/lib/python3.8/site-packages/old
# $ pip install docarray
from old import Document as OldDocument
from old import DocumentArray as OldDocumentArray


def _convert_doc(old: OldDocument) -> Document:
    new = Document(id=old.id, mime_type=old.mime_type)
    if old.tags['content_type'] == 'image':
        new.uri = old.uri
        new.tensor = old.blob
    else:
        new.text = old.text
    for k, v in old.tags.items():
        new.tags[k] = v
    new.embedding = old.embedding
    return new


def convert_dataset(dataset: str, num_workers: int = 8):
    path = f'{dataset}.bin'

    print(f'===> {dataset}')
    print(f'  Loading {dataset} dataset from {path} ...')
    old_docs = OldDocumentArray.load_binary(path)
    print(f'  Old dataset size: {len(old_docs)}')

    print('  Converting docs ...')
    with mp.Pool(processes=num_workers) as pool:
        new_docs = list(tqdm(pool.imap(_convert_doc, old_docs)))

    new_docs = DocumentArray(new_docs)

    print(f'  New dataset size: {len(new_docs)}')

    print('  Saving new docs ...')
    out = f'new.{dataset}.bin'
    new_docs.save_binary(out, compress='gzip')
    print(f'  Saved new docs to {out} ...')


def main():
    """
    Main method.
    """
    datasets = [
        'deepfashion',
        'deepfashion.img10',
        'deepfashion.txt10',
        'deepfashion.ViT-B32',
        'deepfashion.ViT-B16',
        'deepfashion.ViT-L14',
        'nih-chest-xrays',
        'nih-chest-xrays.img10',
        'nih-chest-xrays.txt10',
        'nih-chest-xrays.ViT-B32',
        'nih-chest-xrays.ViT-B16',
        'nih-chest-xrays.ViT-L14',
        'geolocation-geoguessr',
        'geolocation-geoguessr.img10',
        'geolocation-geoguessr.txt10',
        'geolocation-geoguessr.ViT-B32',
        'geolocation-geoguessr.ViT-B16',
        'geolocation-geoguessr.ViT-L14',
        'stanford-cars',
        'stanford-cars.img10',
        'stanford-cars.txt10',
        'stanford-cars.ViT-B32',
        'stanford-cars.ViT-B16',
        'stanford-cars.ViT-L14',
        'bird-species',
        'bird-species.img10',
        'bird-species.txt10',
        'bird-species.ViT-B32',
        'bird-species.ViT-B16',
        'bird-species.ViT-L14',
        'best-artworks',
        'best-artworks.img10',
        'best-artworks.txt10',
        'best-artworks.ViT-B32',
        'best-artworks.ViT-B16',
        'best-artworks.ViT-L14',
    ]
    num_workers = 8

    for dataset in datasets:
        convert_dataset(dataset, num_workers)


if __name__ == '__main__':
    main()



import os

from yaspin.core import Yaspin


def yaspin_extended(*args, **kwargs):
    return YaspinExtended(*args, **kwargs)


class YaspinExtended(Yaspin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __enter__(self):
        if 'NOW_CI_RUN' in os.environ:
            return self
        else:
            return super().__enter__()

    def __exit__(self, exc_type, exc_val, traceback):
        if 'NOW_CI_RUN' not in os.environ:
            return super().__exit__(exc_type, exc_val, traceback)



from now.data_loading.data_loading import load_data
from now.deployment.flow import deploy_flow
from now.dialog import UserInput
from now.finetuning.run_finetuning import finetune_now
from now.finetuning.settings import parse_finetune_settings


def run(user_input: UserInput, tmpdir, kubectl_path: str):
    """
    TODO: Write docs

    :param user_input:
    :param tmpdir:
    :param kubectl_path:
    :return:
    """
    dataset = load_data(user_input)

    finetune_settings = parse_finetune_settings(user_input, dataset)
    if finetune_settings.perform_finetuning:
        print(f'🔧 Perform finetuning!')
        finetune_settings.finetuned_model_name = finetune_now(
            user_input, dataset, finetune_settings, kubectl_path
        )
    (
        gateway_host,
        gateway_port,
        gateway_host_internal,
        gateway_port_internal,
    ) = deploy_flow(
        user_input=user_input,
        finetune_settings=finetune_settings,
        index=dataset,
        tmpdir=tmpdir,
        kubectl_path=kubectl_path,
    )
    return gateway_host, gateway_port, gateway_host_internal, gateway_port_internal



import base64

from docarray import Document
from fastapi import HTTPException


def process_query(text: str, image: str) -> Document:
    if text is None and image is None:
        raise ValueError('Please set one of the value - `image` or `text`')
    if text is not None and image is not None:
        raise ValueError('Please set either image or text not both!')
    try:
        if text is not None:
            query_doc = Document(text=text)
        else:
            base64_bytes = image.encode('utf-8')
            message_bytes = base64.decodebytes(base64_bytes)
            query_doc = Document(blob=message_bytes)
    except BaseException as e:
        raise HTTPException(
            status_code=500,
            detail=f'Not a correct encoded query. Please see the error stack for more information. \n{e}',
        )
    return query_doc



""" This module is the entry point to the finetuning package."""
import os
import tempfile
import warnings
from contextlib import contextmanager
from copy import deepcopy
from os.path import join as osp

import finetuner
from docarray import DocumentArray
from docarray.math.evaluation import ndcg_at_k
from finetuner.tuner.callback import (
    BestModelCheckpoint,
    EarlyStopping,
    EvaluationCallback,
)
from finetuner.tuner.pytorch.losses import TripletLoss
from finetuner.tuner.pytorch.miner import TripletEasyHardMiner
from yaspin import yaspin

from now.constants import PRE_TRAINED_LINEAR_HEADS_MUSIC, Modalities
from now.dialog import UserInput
from now.finetuning.dataset import FinetuneDataset, build_finetuning_dataset
from now.finetuning.embeddings import embed_now
from now.finetuning.settings import FinetuneSettings
from now.hub.head_encoder.head_encoder import LinearHead, get_bi_modal_embedding
from now.hub.hub import push_to_hub
from now.improvements.improvements import show_improvement
from now.utils import sigmap

_BASE_SAVE_DIR = 'now/hub/head_encoder'


def finetune_now(
    user_input: UserInput,
    dataset: DocumentArray,
    finetune_settings: FinetuneSettings,
    kubectl_path: str,
):
    """
    Performs the finetuning procedure:
     1. If embeddings are not present -> compute them using a k8s deployed flow
     2. If bi-modal, prepare the embeddings by concatenating zeros for the opposing modality
     3. Create model and run finetuning, get path to the tuned model and return

    Note, for music we use cached models because the datasets are too large and consume too much time

    :param user_input: The configured user input object
    :param dataset: The dataset with the finetuning labels on all documents. Embeddings are optional and can
        be computed on the fly
    :param finetune_settings: Mainly parameter configuration for the finetuner.fit
    :param kubectl_path: Path to the kubectl binary on the system

    :return: Path to the tuned model.
    """
    if user_input.output_modality == Modalities.MUSIC:
        print(f'⚡️ Using cached hub model for speed')
        return PRE_TRAINED_LINEAR_HEADS_MUSIC[user_input.data]
    dataset = _maybe_add_embeddings(user_input, dataset, kubectl_path)

    dataset = dataset.shuffle(42)

    if finetune_settings.bi_modal:
        _prepare_dataset_bi_modal(dataset)

    finetune_ds = build_finetuning_dataset(dataset, finetune_settings)

    with _finetune_dir() as save_dir:

        finetuned_model_path = _finetune_layer(finetune_ds, finetune_settings, save_dir)

        if (
            "NOW_CI_RUN" not in os.environ
            and user_input.output_modality == Modalities.IMAGE
        ):
            _show_finetune_improvements(
                user_input, finetune_settings, finetune_ds, finetuned_model_path
            )

        executor_name = push_to_hub(save_dir)
    return executor_name


def _finetune_layer(
    finetune_ds: FinetuneDataset, finetune_settings: FinetuneSettings, save_dir: str
) -> str:
    for ds_name, ds in finetune_ds.as_dict().items():
        for doc in ds:
            doc.tensor = doc.embedding
            doc.embedding = None

    assert all([d.embedding is not None for d in finetune_ds.index])

    save_dir = os.path.join(save_dir, 'now', 'hub', 'head_encoder')
    os.makedirs(save_dir, exist_ok=True)

    callbacks = [
        EvaluationCallback(
            finetune_ds.val_query,
            finetune_ds.val_index,
            limit=finetune_settings.eval_match_limit,
            num_workers=8,
            metrics={'ndcg': (ndcg_at_k, {})},
        ),
        BestModelCheckpoint(monitor='ndcg', save_dir=save_dir),
        EarlyStopping(
            monitor='ndcg',
            verbose=False,
            patience=finetune_settings.early_stopping_patience,
        ),
    ]

    print('💪 fine-tuning:')
    input_size = (
        finetune_settings.pre_trained_embedding_size
        if not finetune_settings.bi_modal
        else finetune_settings.pre_trained_embedding_size * 2
    )
    head = LinearHead(input_size, finetune_settings.finetune_layer_size)

    finetuner.fit(
        head,
        train_data=finetune_ds.train,
        eval_data=finetune_ds.val,
        epochs=finetune_settings.epochs,
        learning_rate=finetune_settings.learning_rate,
        batch_size=finetune_settings.batch_size,
        loss=TripletLoss(
            miner=TripletEasyHardMiner(
                pos_strategy=finetune_settings.pos_mining_strat,
                neg_strategy=finetune_settings.neg_mining_strat,
            ),
        ),
        num_items_per_class=finetune_settings.num_items_per_class,
        callbacks=callbacks,
    )
    print('🧠 Perfect! Early stopping triggered since accuracy is great already')

    return os.path.join(save_dir, 'best_model_ndcg')


@contextmanager
def _finetune_dir() -> str:
    with tempfile.TemporaryDirectory() as tmpdir:
        full_save_path = osp(tmpdir, _BASE_SAVE_DIR)
        os.makedirs(full_save_path, exist_ok=True)
        yield full_save_path


def _maybe_add_embeddings(
    user_input: UserInput, dataset: DocumentArray, kubectl_path: str
):
    with yaspin(
        sigmap=sigmap, text="Check if embeddings already exist", color="green"
    ) as spinner:
        if all([d.embedding is not None for d in dataset]):
            spinner.ok('👍')
            return dataset
        else:
            spinner.fail('👎')

    embed_now(user_input, dataset, kubectl_path=kubectl_path)

    assert all([d.embedding is not None for d in dataset]), (
        "Some docs slipped through and" " still have no embedding..."
    )


def _prepare_dataset_bi_modal(dataset: DocumentArray):
    for doc in dataset:
        doc.embedding = get_bi_modal_embedding(doc)


def _show_finetune_improvements(
    user_input: UserInput,
    finetune_settings: FinetuneSettings,
    finetune_ds: FinetuneDataset,
    finetuned_model_path: str,
):
    def restore_content_attribute():
        for doc in finetune_ds.val:
            index_doc = finetune_ds.index[doc.id]
            if index_doc.text:
                doc.text = index_doc.text
            elif index_doc.blob:
                doc.blob = index_doc.blob

    restore_content_attribute()
    val_index_image = deepcopy(DocumentArray(d for d in finetune_ds.val if d.blob))
    val_query_image = deepcopy(
        val_index_image.sample(k=finetune_settings.num_val_queries, seed=42)
    )
    with yaspin(sigmap=sigmap, text="Create overview", color="green") as spinner:
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                show_improvement(
                    user_input.data,
                    user_input.quality,
                    val_query_image,
                    val_index_image,
                    finetune_ds.val_query,
                    finetune_ds.val_index,
                    finetune_settings.pre_trained_embedding_size,
                    finetune_settings.finetune_layer_size,
                    finetuned_model_path,
                    class_label='finetuner_label',
                )
        except Exception as e:
            pass
        spinner.ok('🖼')
        print(
            f'before-after comparison result is saved in the current working directory as image'
        )



from typing import List


class Modalities:
    IMAGE = 'image'
    MUSIC = 'music'
    TEXT = 'text'

    @classmethod
    def as_list(cls) -> List[str]:
        return [cls.IMAGE, cls.MUSIC, cls.TEXT]


class DatasetTypes:
    DEMO = 'demo'
    PATH = 'path'
    URL = 'url'
    DOCARRAY = 'docarray'

    @classmethod
    def as_list(cls) -> List[str]:
        return [cls.DEMO, cls.PATH, cls.URL, cls.DOCARRAY]


class Qualities:
    MEDIUM = 'medium'
    GOOD = 'good'
    EXCELLENT = 'excellent'

    @classmethod
    def as_list(cls) -> List[str]:
        return [cls.MEDIUM, cls.GOOD, cls.EXCELLENT]


BASE_STORAGE_URL = (
    'https://storage.googleapis.com/jina-fashion-data/data/one-line/datasets'
)

IMAGE_MODEL_QUALITY_MAP = {
    Qualities.MEDIUM: ('ViT-B32', 'openai/clip-vit-base-patch32'),
    Qualities.GOOD: ('ViT-B16', 'openai/clip-vit-base-patch16'),
    Qualities.EXCELLENT: ('ViT-L14', 'openai/clip-vit-large-patch14'),
}


class DemoDatasets:
    BEST_ARTWORKS = 'best-artworks'
    NFT_MONKEY = 'nft-monkey'
    TLL = 'tll'
    BIRD_SPECIES = 'bird-species'
    STANFORD_CARS = 'stanford-cars'
    DEEP_FASHION = 'deepfashion'
    NIH_CHEST_XRAYS = 'nih-chest-xrays'
    GEOLOCATION_GEOGUESSR = 'geolocation-geoguessr'
    MUSIC_GENRES_MID = 'music-genres-mid'
    MUSIC_GENRES_LARGE = 'music-genres-large'
    MUSIC_GENRES_EXTRA_LARGE = 'music-genres-xl'
    MUSIC_MID = 'music-mid'
    MUSIC_LARGE = 'music-large'
    MUSIC_EXTRA_LARGE = 'music-xl'
    ROCK_LYRICS = 'rock-lyrics'
    POP_LYRICS = 'pop-lyrics'
    RAP_LYRICS = 'rap-lyrics'
    INDIE_LYRICS = 'indie-lyrics'
    METAL_LYRICS = 'metal-lyrics'


AVAILABLE_DATASET = {
    Modalities.IMAGE: [
        DemoDatasets.BEST_ARTWORKS,
        DemoDatasets.NFT_MONKEY,
        DemoDatasets.TLL,
        DemoDatasets.BIRD_SPECIES,
        DemoDatasets.STANFORD_CARS,
        DemoDatasets.DEEP_FASHION,
        DemoDatasets.NIH_CHEST_XRAYS,
        DemoDatasets.GEOLOCATION_GEOGUESSR,
    ],
    Modalities.MUSIC: [
        DemoDatasets.MUSIC_GENRES_MID,
        DemoDatasets.MUSIC_GENRES_LARGE,
    ],
    Modalities.TEXT: [
        DemoDatasets.ROCK_LYRICS,
        DemoDatasets.POP_LYRICS,
        DemoDatasets.RAP_LYRICS,
        DemoDatasets.INDIE_LYRICS,
        DemoDatasets.METAL_LYRICS,
    ],
}

JC_SECRET = '~/.cache/jina-now/wolf.json'

SURVEY_LINK = 'https://docs.google.com/forms/d/e/1FAIpQLSex5gMN4wuQc63TriwRqREBfdijwOrATPe7RotcPaT1SSPfEw/viewform?usp=pp_url&entry.1126738320=Jina+NOW+pip+package'


PRE_TRAINED_LINEAR_HEADS_MUSIC = {
    DemoDatasets.MUSIC_GENRES_MID: 'FineTunedLinearHeadEncoder:93ea59dbd1ee3fe0bdc44252c6e86a87/'
    'linear_head_encoder_music_2k',
    DemoDatasets.MUSIC_GENRES_LARGE: 'FineTunedLinearHeadEncoder:93ea59dbd1ee3fe0bdc44252c6e86a87/'
    'linear_head_encoder_music_10k',
    DemoDatasets.MUSIC_GENRES_EXTRA_LARGE: 'FineTunedLinearHeadEncoder:93ea59dbd1ee3fe0bdc44252c6e86a87/'
    'linear_head_encoder_music_40k',
}



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field

_ProtoValueType = Optional[Union[bool, float, str, list, dict]]
_StructValueType = Union[
    _ProtoValueType, List[_ProtoValueType], Dict[str, _ProtoValueType]
]


class _NamedScore(BaseModel):
    value: Optional[float] = None


class BaseRequestModel(BaseModel):
    host: str = Field(
        default='localhost', description='Host address returned by the flow deployment.'
    )
    port: int = Field(default=31080, description='Port at which to connect.')

    class Config:
        allow_mutation = False
        case_sensitive = False
        arbitrary_types_allowed = True
